package com.mt.wms.empty.task;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.mt.wms.core.dal.entity.*;
import com.mt.wms.core.dal.service.*;
import com.mt.wms.core.utils.HttpClient;
import com.mt.wms.core.vo.R;
import com.mt.wms.empty.websocket.WebSocketServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.io.IOException;
import java.time.LocalDateTime;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.bouncycastle.asn1.x500.style.RFC4519Style.l;

/**
 * @Author: liguanghao
 * @Date: 2021/11/15 21:55
 * @Version 1.0
 */
@Component
@Transactional
public class RunTaskUtils {
    private Logger logger = LoggerFactory.getLogger(getClass());
    @Resource
    private CurrTaskServiceBiz currTaskServiceBiz;
    @Resource
    private LocationServiceBiz locationServiceBiz;
    @Resource
    private AsynRunTaskService asynRunTaskService;
    @Resource
    private VehicleInfoServiceBiz vehicleInfoServiceBiz;
    @Resource
    private TaskDistanceUtils taskDistanceUtils;
    @Resource
    private PointInfoServiceBiz pointInfoServiceBiz;
    @Resource
    private AutoExeTaskServiceBiz autoExeTaskServiceBiz;
    @Resource
    private KilnInfoServiceBiz kilnInfoServiceBiz;
    @Resource
    private PlcNameSpaceServiceBiz plcNameSpaceServiceBiz;
    @Resource
    private WebSocketServer webSocketServer;

    /**
     * 进炉加工
     * @param currTaskId 任务id
     * @return R
     * @throws InterruptedException
     */
    public R runTaskForInKiln(Long currTaskId) throws InterruptedException, IOException {
        CurrTask currTask = currTaskServiceBiz.getById(currTaskId);
        //起点,起点和终点是液压台code、窑炉code、缓存区code
        String startPosition = currTask.getStartPosition();
        PointInfo startPoint = pointInfoServiceBiz.getOne(new QueryWrapper<PointInfo>().eq(PointInfo.NOTE, startPosition));
        // 可以根据点位类型判断起点是否为液压台
        Integer startPointType = startPoint.getType();
        if (startPointType==3){
            Map<String, Object> yeya = new HashMap();
            //type=1为 入
            yeya.put("type", 1);
            yeya.put("number", startPoint.getCode());
            String yeyaJudgeStatus = JSON.toJSONString(yeya);
            //判断起点液压台是否已放货且提升到位
            String yeyaStatus=null;
            try {
                yeyaStatus = HttpClient.httpPost("http://localhost:8009/yeya/isReadyForYeyaStatus", yeyaJudgeStatus);
            } catch (IOException e) {
                e.printStackTrace();
                logger.error(new Date() + "WCS接口超时未响应！");
            }
            //taskStatus==3满足条件，其他状态值说明液压台不满足开始任务的条件（无货物或者有货物但是没提升到位）
            if (Integer.parseInt(JSONObject.parseObject(yeyaStatus).get("taskStatus").toString()) != 3){
                return R.failed(startPoint.getName()+"号液压台不满足任务执行所必须的条件，请检查液压台是否有货物且提升到位！");
            }
        }
        //是否有小车在运行
        List<VehicleInfo> vehicleInfoList = vehicleInfoServiceBiz.list(new QueryWrapper<VehicleInfo>().eq(VehicleInfo.STATUS, 1));
        if (vehicleInfoList.size()>0){
            //有小车在运行，提示当前有车辆在运行，请稍后
            return R.failed("当前轨道有RGV车辆在运动，请等待当前RGV车辆执行完毕再执行任务！");
        }
        //  2022/2/27  传入任务id,起始点，若终点是窑炉，获取窑炉状态,能放两坨货的窑炉怎么判定状态？(AllowIn)
        KilnInfo kilnInfo = kilnInfoServiceBiz.getById(currTask.getKilnId());
        String kilnNamePlace = plcNameSpaceServiceBiz.getOne(new QueryWrapper<PlcNameSpace>()
                .eq(PlcNameSpace.EQ_TYPE, 0)
                .eq(PlcNameSpace.EQ_ID, kilnInfo.getId())).getName();
        String result = readPlcToString(kilnNamePlace, "AllowIn");
        Boolean resultBoolean = Boolean.valueOf(JSONObject.parseObject(result).get("result").toString());
        //窑炉状态为不可用
        if (kilnInfo.getStatus() == 1) {
            return R.failed("目标窑炉不可用！");
        }
        //窑炉可用未满
        if (resultBoolean){
            //计算出路径最短的车辆id
            Long vehicleId = taskDistanceUtils.chooseVehicle(currTask.getStartPosition(), currTask.getTargetPosition());
            //占用车辆
            VehicleInfo vehicleInfo = vehicleInfoServiceBiz.getById(vehicleId);
            vehicleInfo.setStatus(1);
            vehicleInfo.setUpdateTime(LocalDateTime.now());
            vehicleInfoServiceBiz.updateById(vehicleInfo);
            //更改任务状态为执行中
            currTask.setStatus(1);
            currTask.setUpdateTime(LocalDateTime.now());
            currTaskServiceBiz.updateById(currTask);
            //异步调用车辆
            asynRunTaskService.asynRunTask(currTask.getId(),vehicleId);
            return R.ok("操作成功，任务已开始执行。");
        } else {
            //窑炉已满
            //查询缓存区库位是否已满
            int count = locationServiceBiz.count(new QueryWrapper<Location>()
                    .eq(Location.STATUS, 0)
                    .eq(Location.VALID, 1));
            //缓存区未满
            if (count >0) {
                //分配一个缓存区库位
                Location location = locationServiceBiz.list(new QueryWrapper<Location>()
                        .eq(Location.STATUS, 1)
                        .eq(Location.VALID, 1)).get(0);
                //更新 location 表对应库位状态为占用
                location.setStatus(1);
                location.setUpdateTime(LocalDateTime.now());
                locationServiceBiz.updateById(location);
                //通过库位code去point_info表中查出该库位对应的点
                PointInfo pointInfo = pointInfoServiceBiz.getOne(new QueryWrapper<PointInfo>().eq(PointInfo.NOTE, location.getCode()));
                //计算出路径最短的车辆id,起点液压台，终点缓存区库位点
                Long vehicleId = taskDistanceUtils.chooseVehicle(currTask.getStartPosition(), location.getCode());
                //占用车辆
                VehicleInfo vehicleInfo = vehicleInfoServiceBiz.getById(vehicleId);
                vehicleInfo.setStatus(1);
                vehicleInfo.setUpdateTime(LocalDateTime.now());
                vehicleInfoServiceBiz.updateById(vehicleInfo);
                //更改任务为入缓存区
                currTask.setIsCache(1);
                currTask.setCacheSatatus(1);
                currTask.setUpdateTime(LocalDateTime.now());
                currTask.setLocationId(location.getId());
                currTaskServiceBiz.updateById(currTask);
                //异步调用车辆
                asynRunTaskService.asynRunTaskToWarehouse(currTask.getId(),vehicleId);
                return R.ok("操作成功！当前目标窑炉已满，托盘准备加入缓存区待加工队列。");
            }
            return R.failed("当前目标窑炉已满！缓存区已满！请稍后重试!");
        }
    }

    /**
     * 加工完成、出炉。传入窑炉id，查询正在该窑炉加工的currTask,查询是否有空闲车辆，有，
     * （查询是否为自动任务，是，查询下一步的窑炉是否空闲，是，调车去下一窑炉加工，否，暂存缓存区）
     * 呼叫车辆暂存缓存区，车辆调度任务结束后，（结束后再查询缓R存区队列避免车辆路线冲突）查询该窑炉的缓存区队列，开始下一托盘货物调度进窑炉
     * @return R
     */
    public R runTaskForOutKiln(Long kilnId) throws IOException {
        //查询正在该窑炉加工的currTask
        CurrTask currTask = currTaskServiceBiz.getOne(new QueryWrapper<CurrTask>()
                .eq("is_in", 1)
                .eq("status", 1)
                .eq("kiln_id", kilnId));
        //查询是否为自动任务
        if (currTask.getIsAuto()==1) {
            //查询自动任务的哪一步的设备id是该窑炉
            int firstCount = autoExeTaskServiceBiz.count(new QueryWrapper<AutoExeTask>().eq(AutoExeTask.F_TASK_ID, currTask.getId()));
            if (firstCount == 1) {
                //当前为自动任务第一步
                AutoExeTask autoExeTask = autoExeTaskServiceBiz.getOne(new QueryWrapper<AutoExeTask>().eq(AutoExeTask.F_TASK_ID, currTask.getId()));
                //第二部的currTaskId
                Long taskId = autoExeTask.getTTaskId();
                CurrTask currTaskTwo = currTaskServiceBiz.getById(taskId);
                try {
                    logger.info("自动任务第一步:"+currTask.getTaskCode()+"已执行完成。开始执行第二步任务:"+currTaskTwo);
                    //  2022/2/27 websocket推送到页面显示
                    webSocketServer.sendtoAll("自动任务第一步:"+currTask.getTaskCode()+"已执行完成。开始执行第二步任务:"+currTaskTwo);
                     return runTaskForInKiln(taskId);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            } else {
                //当前加工步骤不为第一步
                int secondCount = autoExeTaskServiceBiz.count(new QueryWrapper<AutoExeTask>().eq(AutoExeTask.T_TASK_ID, currTask.getId()));
                if (secondCount == 2) {
                    //当前为自动任务第二步
                    AutoExeTask autoExeTask = autoExeTaskServiceBiz.getOne(new QueryWrapper<AutoExeTask>().eq(AutoExeTask.T_TASK_ID, currTask.getId()));
                    //该自动任务的工艺类型为2，还有第三步
                    if (autoExeTask.getProcessFlowType()==2){
                        Long thTaskId = autoExeTask.getThTaskId();
                        CurrTask currTaskThree = currTaskServiceBiz.getById(thTaskId);
                        try {
                            logger.info("自动任务第二步:"+currTask.getTaskCode()+"已执行完成。开始执行第三步任务:"+currTaskThree);
                            // 2022/2/27 websocket推送到页面显示
                            webSocketServer.sendtoAll("自动任务第二步:"+currTask.getTaskCode()+"已执行完成。开始执行第三步任务:"+currTaskThree);
                            return runTaskForInKiln(thTaskId);
                        } catch (InterruptedException e) {
                            e.printStackTrace();
                        }
                    }

                }

            }
        }
        //查询缓存区库位是否已满
        int count = locationServiceBiz.count(new QueryWrapper<Location>()
                .eq(Location.STATUS, 0)
                .eq(Location.VALID, 1));
        //缓存区未满
        if (count >0) {
            List<Location> locationList = locationServiceBiz.list(new QueryWrapper<Location>()
                    .eq(Location.STATUS, 0)
                    .eq(Location.VALID, 1));
            //是否有小车在运行
            List<VehicleInfo> vehicleInfoList = vehicleInfoServiceBiz.list(new QueryWrapper<VehicleInfo>().eq(VehicleInfo.STATUS, 1));
            if (vehicleInfoList.size()>0){
                //有小车在运行，提示当前有车辆在运行，请稍后
                return R.failed("当前轨道有RGV车辆在运动，请等待当前RGV车辆执行完毕再执行任务！");
            }
            //计算出路径最短的车辆id
            KilnInfo kilnInfo = kilnInfoServiceBiz.getById(kilnId);
            //窑炉为起点,缓存区为终点
            Location location = locationList.get(0);
            Long vehicleId = taskDistanceUtils.chooseVehicle(kilnInfo.getCode(), location.getCode());
            //占用车辆
            VehicleInfo vehicleInfo = vehicleInfoServiceBiz.getById(vehicleId);
            vehicleInfo.setStatus(1);
            vehicleInfo.setUpdateTime(LocalDateTime.now());
            vehicleInfoServiceBiz.updateById(vehicleInfo);
            //修改任务状态
            currTask.setIsOut(0);
            currTask.setTaskType(2);
            currTask.setUpdateTime(LocalDateTime.now());
            currTask.setOutStartPosition(kilnInfo.getCode());
            currTask.setOutTargetPosition(location.getCode());
            currTaskServiceBiz.updateById(currTask);
            //  2022/2/16 异步调用车辆
            asynRunTaskService.asynRunTaskForKilnToWarehouse(currTask.getId(),vehicleInfoList.get(0).getId());
            return R.ok("操作成功！当前目标窑炉已满，托盘加入缓存区待加工队列。");
        }else {
            return R.ok("当前缓存区已满，无法从加工炉出炉到缓存区。");
        }
    }

    /**
     * 从缓存区出库
     * @param locationId 库位id
     * @param yeyaTaiCode 液压台code
     * @return R
     */
    public R runTaskForOut(Long locationId,String yeyaTaiCode) throws IOException {
        Location location = locationServiceBiz.getById(locationId);
        //是否有小车在运行
        List<VehicleInfo> vehicleInfoList = vehicleInfoServiceBiz.list(new QueryWrapper<VehicleInfo>().eq(VehicleInfo.STATUS, 1));
        if (vehicleInfoList.size()>0){
            //有小车在运行，提示当前有车辆在运行，请稍后
            return R.failed("当前轨道有RGV车辆在运动，请等待当前RGV车辆执行完毕再执行任务！");
        }
        //计算出路径最短的车辆id
        Long vehicleId = taskDistanceUtils.chooseVehicle(location.getCode(), yeyaTaiCode);
        //占用车辆
        VehicleInfo vehicleInfo = vehicleInfoServiceBiz.getById(vehicleId);
        vehicleInfo.setStatus(1);
        vehicleInfo.setUpdateTime(LocalDateTime.now());
        vehicleInfoServiceBiz.updateById(vehicleInfo);
        //异步调用车辆
        asynRunTaskService.asynRunOutStock(locationId,yeyaTaiCode,vehicleInfo.getId());
        return R.ok("操作成功，任务已开始执行。");
    }

    /**
     * 获取变量值
     * @param nameSpace nameSpace前缀
     * @param variableName 变量名称
     * @return result字符串
     */
    private String readPlcToString(String nameSpace, String variableName){
        String identifier=nameSpace.concat(variableName);
        Map<String, Object> json = new HashMap();
        json.put("nameSpace", 6);
        json.put("plcName", "plc1");
        json.put("identifier",identifier);
        String result = null;
        try {
            result = HttpClient.httpPost("http://localhost:8009/opcua/read", JSON.toJSONString(json));
        } catch (IOException e) {
            e.printStackTrace();
            logger.info("readPLC： "+identifier+" 的值失败");
        }
        return result;
    }

}
